@@ -1,17 +1,17 @@
-/*	This is pstree written by Fred Hucht (c) 1993-2004	*
+/*	This is pstree written by Fred Hucht (c) 1993-2010	*
  *	EMail: fred AT thp.Uni-Duisburg.de			*
  *	Feel free to copy and redistribute in terms of the	*
  * 	GNU public license. 					*
  *
- * $Id: pstree.c,v 2.27 2005-04-08 22:08:45+02 fred Exp $
+ * $Id: pstree.c,v 2.33 2009-11-10 22:12:39+01 fred Exp $
  */
 static char *WhatString[]= {
-  "@(#)pstree $Revision: 2.27 $ by Fred Hucht (C) 1993-2004",
+  "@(#)pstree $Revision: 2.33 $ by Fred Hucht (C) 1993-2007",
   "@(#)EMail: fred AT thp.Uni-Duisburg.de",
-  "$Id: pstree.c,v 2.27 2005-04-08 22:08:45+02 fred Exp $"
+  "$Id: pstree.c,v 2.33 2009-11-10 22:12:39+01 fred Exp $"
 };
 
-#define MAXLINE 512
+#define MAXLINE 8192
 
 #if defined(_AIX) || defined(___AIX)	/* AIX >= 3.1 */
 /* Under AIX, we directly read the process table from the kernel */
@@ -48,7 +48,7 @@ extern getargs(struct ProcInfo *, int, char *, int);
 #  define PSVARS	&P[i].uid, &P[i].pid, &P[i].ppid, &P[i].pgid, &P[i].thcount, P[i].cmd
 #  define PSVARSN	6
 /************************************************************************/
-#elif defined(__linux) || (defined __alpha && defined(_SYSTYPE_BSD))
+#elif defined(__linux) || (defined __alpha && defined(_SYSTYPE_BSD) || defined (Tru64))
 /* TRU64 contributed by Frank Parkin <fparki AT acxiom.co.uk>
  */
 #  ifdef __linux
@@ -135,9 +135,16 @@ extern getargs(struct ProcInfo *, int, char *, int);
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>		/* For str...() */
+#ifdef NEED_SNPRINTF
+#include <stdarg.h>
+int snprintf(char *, int, char *, ...);
+#endif
 #include <unistd.h>		/* For getopt() */
 #include <pwd.h>		/* For getpwnam() */
 
+#include <sys/ioctl.h>		/* For TIOCGSIZE/TIOCGWINSZ */
+/* #include <termios.h> */
+
 #ifdef DEBUG
 # include <errno.h>
 #endif
@@ -169,17 +176,26 @@ struct TreeChars {
  * | \-+=   01112 ...    B LPPG 01112 ...
  * |   |--= 01113 ...    B   CSSG 01113 ...
  * |   \--= 01114 ...    B   LSSG 01114 ...
- * \-+- 01115 ...        LSSN 01115 ...
+ * \--- 01115 ...        LSSN 01115 ...
  */
 
-enum { G_ASCII = 0, G_PC850 = 1, G_VT100 = 2, G_LAST };
+enum { G_ASCII = 0, G_PC850 = 1, G_VT100 = 2, G_UTF8 = 3, G_LAST };
 
 /* VT sequences contributed by Randall Hopper <rhh AT ct.picker.com> */
+/* UTF8 sequences contributed by Mark-Andre Hopf <mhopf AT mark13.org> */
 static struct TreeChars TreeChars[] = {
   /* SS          PP          G       N       C       B       L      sg      eg      init */
   { "--",       "-+",       "=",    "-",    "|",    "|",    "\\",   "",     "",     ""             }, /*Ascii*/
   { "\304\304", "\304\302", "\372", "\304", "\303", "\263", "\300", "",     "",     ""             }, /*Pc850*/
-  { "qq",       "qw",       "`",    "q",    "t",    "x",    "m",    "\016", "\017", "\033(B\033)0" }  /*Vt100*/
+  { "qq",       "qw",       "`",    "q",    "t",    "x",    "m",    "\016", "\017", "\033(B\033)0" }, /*Vt100*/
+  { "\342\224\200\342\224\200",
+    /**/        "\342\224\200\342\224\254",
+    /**/                    "=",
+    /**/                            "\342\224\200",
+    /**/                                    "\342\224\234",
+    /**/                                            "\342\224\202",
+    /**/                                                    "\342\224\224",
+    /**/                                                            "",     "",     ""             }  /*UTF8*/
 }, *C;
 
 int MyPid, NProc, Columns, RootPid;
@@ -188,13 +204,16 @@ char *name = "", *str = NULL, *Progname;
 long ipid = -1;
 char *input = NULL;
 
+int atLdepth=0;    /* LOPTION - track how deep in the print chain we are */
+int maxLdepth=100; /* LOPTION - will be changed by -l n option */
+
 #ifdef DEBUG
 int debug = FALSE;
 #endif
 
 struct Proc {
   long uid, pid, ppid, pgid;
-  char name[9], cmd[MAXLINE];
+  char name[32], cmd[MAXLINE];
   int  print;
   long parent, child, sister;
   unsigned long thcount;
@@ -205,10 +224,11 @@ void uid2user(uid_t uid, char *name, int len) {
 #define NUMUN 128
   static struct un_ {
     uid_t uid;
-    char name[9];
+    char name[32];
   } un[NUMUN];
   static short n = 0;
   short i;
+  char uid_name[32];
   char *found;
 #ifdef DEBUG
   if (name == NULL) {
@@ -222,7 +242,13 @@ void uid2user(uid_t uid, char *name, int len) {
     found = un[i].name;
   } else {
     struct passwd *pw = getpwuid(uid);
-    found = pw->pw_name;
+    if (pw) {
+      found = pw->pw_name;
+    } else {
+      /* fix by Stan Sieler & Philippe Torche */
+      snprintf(uid_name, sizeof(uid_name), "#%d", uid);
+      found = uid_name;
+    }
     if (n < NUMUN) {
       un[n].uid = uid;
       strncpy(un[n].name, found, 9);
@@ -515,7 +541,7 @@ int GetProcesses(void) {
 int GetRootPid(void) {
   int me;
   for (me = 0; me < NProc; me++) {
-    if (P[me].pid == -1) return P[me].pid;
+    if (P[me].pid == 1) return P[me].pid;
   }
   /* PID == 1 not found, so we'll take process with PPID == 0
    * Fix for TRU64 TruCluster with uniq PIDs
@@ -530,9 +556,15 @@ int GetRootPid(void) {
   for (me = 0; me < NProc; me++) {
     if (P[me].ppid == 1) return P[me].pid;
   }
+  /* Still nothing. Maybe it is something like Solaris Zone. We'll take
+   * the process with PID == PPID */
+  for (me = 0; me < NProc; me++) {
+    if (P[me].pid == P[me].ppid) return P[me].pid;
+  }
   /* Should not happen */
   fprintf(stderr,
-	  "%s: No process found with PID == 1 || PPID == 0 || PPID == 1, contact author.\n",
+	  "%s: No process found with PID == 1 || PPID == 0 || PPID == 1\n"
+	  "          || PID == PPID, contact author.\n",
 	  Progname);
   exit(1);
 }
@@ -657,6 +689,10 @@ void PrintTree(int idx, const char *head) {
   if (head[0] == '\0' && !P[idx].print) return;
   
   if (P[idx].thcount > 1) snprintf(thread, sizeof(thread), "[%ld]", P[idx].thcount);
+ 
+  if(atLdepth == maxLdepth) return;    /* LOPTION */
+  ++atLdepth;                          /* LOPTION */
+ 
   
   snprintf(out, sizeof(out),
 	   "%s%s%s%s%s%s %05ld %s %s%s" /*" (ch=%d, si=%d, pr=%d)"*/,
@@ -680,6 +716,9 @@ void PrintTree(int idx, const char *head) {
   
   for (child = P[idx].child; EXIST(child); child = P[child].sister)
     PrintTree(child, nhead);
+
+  --atLdepth;                          /* LOPTION */
+
 }
 
 void Usage(void) {
@@ -690,14 +729,15 @@ void Usage(void) {
 #ifdef DEBUG
 	  "[-d] "
 #endif
-	  "[-f file] [-g] [-u user] [-U] [-s string] [-p pid] [-w] [pid ...]\n"
+	  "[-f file] [-g n] [-u user] [-U] [-s string] [-p pid] [-w] [pid ...]\n"
 	  /*"   -a        align output\n"*/
 #ifdef DEBUG
 	  "   -d        print debugging info to stderr\n"
 #endif
 	  "   -f file   read input from <file> (- is stdin) instead of running\n"
 	  "             \"%s\"\n"
-	  "   -g n      use graphics chars for tree. n=1: IBM-850, n=2: VT100\n"
+	  "   -g n      use graphics chars for tree. n=1: IBM-850, n=2: VT100, n=3: UTF-8\n"
+	  "   -l n      print tree to n level deep\n"
 	  "   -u user   show only branches containing processes of <user>\n"
 	  "   -U        don't show branches containing only root processes\n"
           "   -s string show only branches containing process with <string> in commandline\n"
@@ -725,7 +765,7 @@ int main(int argc, char **argv) {
   Progname = strrchr(argv[0],'/');
   Progname = (NULL == Progname) ? argv[0] : Progname + 1;
   
-  while ((ch = getopt(argc, argv, "df:g:hp:s:u:Uw?")) != EOF)
+  while ((ch = getopt(argc, argv, "df:g:hl:p:s:u:Uw?")) != EOF)
     switch(ch) {
       /*case 'a':
 	align   = TRUE;
@@ -747,6 +787,10 @@ int main(int argc, char **argv) {
       }
       C = &TreeChars[graph];
       break;
+    case 'l':                                 /* LOPTION */
+      maxLdepth = atoi(optarg);               /* LOPTION */
+      if(maxLdepth < 1) maxLdepth = 1;        /* LOPTION */
+      break;                                  /* LOPTION */
     case 'p':
       showall = FALSE;
       ipid    = atoi(optarg);
@@ -816,8 +860,16 @@ int main(int argc, char **argv) {
   if (wide)
     Columns = MAXLINE - 1;
   else {
-#ifdef HAS_TERMDEF
+#if defined (HAS_TERMDEF)
     Columns = atoi((char*)termdef(fileno(stdout),'c'));
+#elif defined(TIOCGWINSZ)
+    struct winsize winsize;
+    ioctl(fileno(stdout), TIOCGWINSZ, &winsize);
+    Columns = winsize.ws_col;
+#elif defined(TIOCGSIZE)
+    struct ttysize ttysize;
+    ioctl(fileno(stdout), TIOCGSIZE, &ttysize);
+    Columns = ttysize.ts_cols;
 #else
     char *env = getenv("COLUMNS");
     Columns = env ? atoi(env) : 80;
@@ -831,6 +883,10 @@ int main(int argc, char **argv) {
 
   if (Columns >= MAXLINE) Columns = MAXLINE - 1;
   
+#ifdef DEBUG
+  if (debug) fprintf(stderr, "Columns = %d\n", Columns);
+#endif
+  
   MakeTree();
   MarkProcs();
   DropProcs();
@@ -863,8 +919,43 @@ static char * strstr(s1, s2)
 }
 #endif /* NEED_STRSTR */
 
+#ifdef NEED_SNPRINTF
+/* Contributed by Michael E White */
+int snprintf(char *name, int namesiz, char *format, ...)
+{
+  int retval;
+  char bigbuf[1024] = {'\0'};  
+  va_list ap;
+  va_start(ap, format);
+  retval = vsprintf(bigbuf,format,ap);
+  va_end(ap);
+  if (retval > namesiz) retval = namesiz;
+  strncpy(name, bigbuf, retval);
+  name[retval] = '\0';
+  return retval;
+}
+#endif  /* NEED_SNPRINTF */
+
 /*
  * $Log: pstree.c,v $
+ * Revision 2.33  2009-11-10 22:12:39+01  fred
+ * Added UTF8, enlarged MAXLINE
+ *
+ * Revision 2.32  2007-10-26 21:39:50+02  fred
+ * Added option -l provided by Michael E. White <mewhite AT us.ibm.com>
+ *
+ * Revision 2.31  2007-06-08 17:45:23+02  fred
+ * Fixed problem with users with long login name (Reported by Oleg A. Mamontov)
+ *
+ * Revision 2.30  2007-05-10 23:13:04+02  fred
+ * *** empty log message ***
+ *
+ * Revision 2.29  2007-05-10 22:37:13+02  fred
+ * Added fix for Solaris Zone and bug fix from Philippe Torche
+ *
+ * Revision 2.28  2007-05-10 22:01:07+02  fred
+ * Added new determination of window width
+ *
  * Revision 2.27  2005-04-08 22:08:45+02  fred
  * Also accept PPID==1 if nothing else is found. Should fix problem with
  * FreeBSD and security.bsd.see_other_uids=0.
